# Ajuste no lineal de elipses
# Del paper:
# Sung Joon Ahn, Wolfgang Rauh, Hans-Jürgen Warnecke,
# Least-squares orthogonal distances fitting of circle, sphere,
# ellipse, hyperbola, and parabola
# Pattern Recognition,
# Volume 34, Issue 12, December 2001, Pages 2283-2303
# https://doi.org/10.1016/S0031-3203(00)00152-7
#
# Código en python
# Fraga 11.08.2020
#
import pprime
import numpy as np
import math
import sys

n = len( sys.argv )
if n != 7:
	print( "Args: archivo_puntos a0 b0 Xc0 Yc0 alfa0" )
	sys.exit(1)

# La elipse inicial
Dva = np.zeros( (5,) )
va = np.zeros( (5,) )
va[0] = float(sys.argv[2])
va[1] = float(sys.argv[3])
va[2] = float(sys.argv[4])
va[3] = float(sys.argv[5])
va[4] = float(sys.argv[6])

# Leemos los puntos de entrada
P = np.genfromtxt( sys.argv[1], delimiter=' ' )
# print( P.size )
# print( P.shape )
[n, m] = P.shape

# Para almacenar los puntos locales
R = np.zeros( (2,2) )
Q = np.zeros( (2,2) )
Qinv = np.zeros( (2,2) )
B = np.zeros( (2, 6) )
# Los residuos los almaceno en la última columna de J
J = np.zeros( (2*n, 6) )

# print( n )
# Los puntos deben pasarse al sistema de coordenadas local
i = 0
while i < 20 :
	a = va[0]
	b = va[1]
	a2 = a*a
	b2 = b*b
	c = math.cos( va[4] )
	s = math.sin( va[4] )
	R[0,0] = c; R[0,1] = -s
	R[1,0] = s; R[1,1] =  c
	sigma = 0.0
	j = 0
	while j < n :
		x = P[j][0] - va[2]
		y = P[j][1] - va[3]

		xi = c*x + s*y
		yi =-s*x + c*y
		# print( xi, yi )

		[x, y] = pprime.pprime( va, [xi, yi] )

		# Calcular la jacobiana
		#
		# J_X'i =(R^{-1} Q^{-1} B)|x=x'i
		#         2x2     2x2  2x5
		#            2x2     2x5
		#               2x5
		B[0,0] = a*(b2 - y*y)
		B[1,0] = 2.0*a*y*( xi - x )

		B[0,1] = b*(a2 - x*x )
		B[1,1] = -2.0*b*x*(yi - y)

		B[0,2] = b2*x*c - a2*y*s 
		B[1,2] = b2*(yi - y)*c + a2*(xi - x)*s

		B[0,3] = b2*x*s + a2*y*c
		B[1,3] = b2*(yi - y)*s - a2*(xi - x)*c 

		B[0,4] = (a2 - b2)*x*y
		B[1,4] = (a2 - b2)*(x*x - y*y - x*xi + y*yi)

		Q[0,0] = b2*x; Q[0,1] = a2*y
		Q[1,0] = (a2-b2)*y + b2*yi
		Q[1,1] = (a2-b2)*x - a2*xi

		# Tengo que invertir Q (uso determinantes)
		d = Q[0,0]*Q[1,1] - Q[1,0]*Q[0,1]
		Qinv[0,0] = Q[1,1]/d
		Qinv[0,1] = -Q[0,1]/d
		Qinv[1,0] = -Q[1,0]/d
		Qinv[1,1] = Q[0,0]/d

		J[2*j:2*j+2, : ] = R @ Qinv @ B

		X = c*x - s*y + va[2]
		Y = s*x + c*y + va[3]
		vX = P[j][0] - X
		vY = P[j][1] - Y
		J[2*j, 5] = vX
		J[2*j+1, 5] = vY
		sigma += vX*vX + vY*vY

		j += 1

	# J tiene dimensiones 2*n x 6
	# print( J.shape )
	Result = np.linalg.qr( J, mode='r' );
	# print( Result.shape )
	# print( Result )
	# Result tiene dimensiones 6, 6
	pprime.backSubstitution( Result, 6, Dva )

	norma = np.linalg.norm( Dva )
	va += 1.0*Dva

	if va[0] < va[1] :
		tmp = va[0]
		va[0] = va[1]
		va[1] = tmp
		if va[4] < 0.0 :
			va[4] += math.pi/2
		else :
			va[4] -= math.pi/2

	print( va, norma, math.sqrt(sigma) )

	i += 1

[U,d,VT] = np.linalg.svd( Result[ 0:5, 0:5 ] )
covs = np.zeros( (5,5) )
j = 0
while j < 5: 
	k = 0
	while k < 5: 
		i = 0
		suma = 0.0
		while i < 5: 
			suma += VT[i,j]*VT[i,k]/(d[i]*d[i]) 
			i += 1

		covs[j,k] = suma
		k += 1
	j += 1

print( "sigmas:" )
i = 0
v = sigma/(2*n - 5)
while i < 5 :
	print( math.sqrt(v*covs[i,i]), end=' ' )	
	i += 1
print( )
print( "correlaciones:" )
i = 0
while i < 5 :
	j = 0
	while j <= i :
		rho = covs[i,j]/math.sqrt( covs[i,i] * covs[j,j] )
		print( rho, end=' ' )
		j += 1
	print()
	i += 1
print()
